"""
Project management tools.
"""

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

import os
import shutil
import yaml

import click

from asimov import config
from asimov import storage
from asimov.ledger import Ledger
from asimov import gitlab



def make_project(name, root, working="working", checkouts="checkouts", results="results"):
    """
    Create a new project called NAME.

    This command creates a new asimov project, creating the appropriate
    directory structure, and creating a blank ledger.
    """
    import venv
    import pathlib

    pathlib.Path(root).mkdir(parents=True, exist_ok=True)
    os.chdir(root)

    config.set("project", "name", name)
    config.set("project", "root", root)

    project_name = name

    # Make the virtual environment
    # builder = venv.EnvBuilder(system_site_packages=False,
    #                           clear=False,
    #                           symlinks=False,
    #                           upgrade=False,
    #                           with_pip=True,
    #                           prompt=f"Asimov {project_name}")

    # builder.create("environment")

    config.set("general", "environment", "environment")

    # Make the working directory
    pathlib.Path(working).mkdir(parents=True, exist_ok=True)
    config.set("general", "rundir_default", working)

    # Make the git directory
    pathlib.Path(checkouts).mkdir(parents=True, exist_ok=True)
    config.set("general", "git_default", checkouts)

    # Make the results store
    storage.Store.create(root=results, name=f"{project_name} storage")
    config.set("storage", "directory", results)

    # Make the ledger
    config.set("ledger", "engine", "yamlfile")
    config.set("ledger", "location", "ledger.yml")

    # Set the default environment
    python_loc = shutil.which("python").split("/")[:-2]
    config.set("pipelines", "environment", os.path.join("/", *python_loc))
    
    Ledger.create(engine="yamlfile",
                  name=project_name,
                  location="ledger.yml")

    with open("asimov.conf", "w") as config_file:
        config.write(config_file)


@click.command()
@click.argument("name")
@click.option("--root", default=os.getcwd(),
              help="Location to create the project, default is the current directory.")
@click.option("--working", default="working",
              help="The location to store working directories, default is a directory called 'working' inside the current directory.")
@click.option("--checkouts", default="checkouts",
              help="The location to store cloned git repositories.")
@click.option("--results", default="results",
              help="The location where the results store should be created.")
def init(name, root, working="working", checkouts="checkouts", results="results"):
    """
    Roll-out a new project.
    """
    make_project(name, root, working=working, checkouts=checkouts, results=results)
    click.echo(click.style("●", fg="green") + f" New project created successfully!")

@click.command()
@click.argument("location")
def clone(location):
    import venv
    import pathlib
    import shutil

    working = "working"
    results = "results"
    
    remote_config = os.path.join(location, "asimov.conf")
    config = configparser.ConfigParser()
    config.read([remote_config])
    click.echo(f'Cloning {config.get("project", "name")}')
    root = os.path.join(os.getcwd(), config.get("project", "name").lower().replace(" ", "-"))
    pathlib.Path(root).mkdir(parents=True, exist_ok=True)
    #os.chdir(root)
    config.set("project", "root", root)
    # Make the virtual environment
    #builder = venv.EnvBuilder(system_site_packages=False,
    #                          clear=False,
    #                          symlinks=False,
    #                          upgrade=False,
    #                          with_pip=True,
    #                          prompt=f"Asimov {project_name}")

    #builder.create("environment")

    #config.set("general", "environment", "environment")

    # Make the working directory
    #shutil.copytree(os.path.join(config.get("general", "rundir_default"), working)
    #config.set("general", "rundir_default", working)

    # Make the git directory
    #pathlib.Path(checkouts).mkdir(parents=True, exist_ok=True)
    #config.set("general", "git_default", checkouts)

    # Copy the results store
    #shutil.copyfile(os.path.join(location, config.get("storage", "results_store")), results)
    shutil.copytree(os.path.join(location, config.get("storage", "results_store")), results)
    config.set("storage", "results_store", results)

    # Make the ledger
    if config.get("ledger", "engine") == "yamlfile":
        shutil.copyfile(os.path.join(location, config.get("ledger", "location")), "ledger.yml")
    elif config.get("ledger", "engine") == "gitlab":
        _, repository = connect_gitlab(config)

        events = gitlab.find_events(repository,
                                update=False,
                                subset=[None],
                                label=config.get("gitlab", "event_label"),
                                repo=False)

        total = []
        for event in events:
            total.append(yaml.safe_load(event.event_object.to_yaml()))
        with open("ledger.yml", "w") as f:
            f.write(yaml.dump(total))

    config.set("ledger", "engine", "yamlfile")
    config.set("ledger", "location", "ledger.yml")

    with open("asimov.conf", "w") as config_file:
        config.write(config_file)



import os
import collections
import yaml
import json
from math import floor
import ligo.gracedb
import gwpy
import ast
import gwpy.timeseries
from gwdatafind import find_urls
from gwpy.segments import DataQualityFlag
from glue.lal import Cache
import numpy as np

import click

from git import GitCommandError

from asimov import config
from asimov import current_ledger as ledger
from asimov.event import Production, Event,  DescriptionException
from asimov import gitlab
from asimov.cli import find_calibrations
from asimov.utils import update

@click.group()
def event():
    """
    Commands to handle events & collections.
    """
    pass

@click.option("--search", "-s", "search", default=None, help="Search criteria.")
@click.option("--old", "oldname", default=None, help="The old superevent ID for this event.")
@click.option("--gid", "gid", default=None, help="The GraceDB GID for the event (for legacy events)")
@click.option("--superevent", "superevent", default=None, help="The superevent for the event.")
@click.option("--repository", "repo", default=None, help="The location of the repository for this event.")
@click.option("--name", "-n", "name", default=None, help="The name for the event.")
@event.command()
def create(name=None, oldname=None, gid=None, superevent=None, repo=None, search=None):
    """
    Create a new event record in the ledger.

    Parameters
    ----------
    superevent : str
       The ID of the superevent to be used from GraceDB
    name : str
       The name of the event to be recorded in the issue tracker
    names : path, optional
        The path to the name file which maps between old and new super event IDs
    oldname : str, optional
        The old name of the event.
    search : str, optional
        A string of search criteria
    """
    import pathlib
    
    if gid or superevent or search:
        from ligo.gracedb.rest import GraceDb, HTTPError 
        client = GraceDb(service_url=config.get("gracedb", "url"))
        r = client.ping()

    if search:
        event_iterator = client.events(search)

        for event in event_iterator:

            event = Event(name=name,
                          repository=repo,
                          calibration = {},
                          interferometers=interferometers,
            )
            working_dir = os.path.join(config.get('general', 'rundir_default'), name)

            event.meta['working directory'] = working_dir
            pathlib.Path(working_dir).mkdir(parents=True, exist_ok=True)
            ledger.update_event(event)
        
    if superevent:
        data = client.superevent(superevent).json()
        try:
            event_data = client.event(data['preferred_event']).json()
        except KeyError:
            event_data = data['preferred_event_data']
        
        gid = data['preferred_event_data']['graceid']
        interferometers = event_data['instruments'].split(",")
        if not name:
            name = superevent
    elif gid:
        event_data = client.event(gid).json()
        interferometers = event_data['instruments'].split(",")
        if not name:
            name = gid
    else:
        event_data = None
        interferometers = []

    if gid or superevent:
        event_url = f"{config.get('gracedb', 'url')}/events/{gid}/view/"
        
    if not repo:
        repo = None
        #repo = f"git@git.ligo.org:pe/O3/{name}"


    event = Event(name=name,
                  repository=repo,
                  calibration = {},
                  interferometers=interferometers,
    )

    if oldname:
        event.meta['old superevent'] = oldname
    if gid:
        event.meta['event time'] = event_data['gpstime']
        event.meta['gid'] = gid

    working_dir = os.path.join(config.get('general', 'rundir_default'), name)

    event.meta['working directory'] = working_dir
    pathlib.Path(working_dir).mkdir(parents=True, exist_ok=True)
    ledger.update_event(event)

@click.argument("delete")
@event.command()
def delete(event):
    """
    Delete an event from the ledger.
    """
    ledger.delete_event(event_name=event)
    
# @click.argument("event")
# @click.option("--yaml", "yaml", default=None)
# @click.option("--ini", "ini", default=None)
# @event.command()
# def populate(event, yaml, ini):
#     """
#     Populate an event ledger with data from ini or yaml files.
#     """

#     event = ledger.get_event(event)
#     # Check the calibration files for this event
#     click.echo("Check the calibration.")
#     click.echo(event.name)
#     calibration(event=event.name)
#     # Check the IFOs for this event
#     click.echo("Check the IFO list")
#     try:
#         checkifo(event.name)
#     except:
#         pass

#     if yaml:
#         add_data(event.name, yaml)

@click.argument("event", default=None)
@click.option("--json", "json_data", default=None)
@event.command()
def configurator(event, json_data=None):    
    """Add data from the configurator."""
    event = ledger.get_event(event)
    if json_data:
        with open(json_data, "r") as datafile:
            data = json.load(datafile)

    new_data = {"quality": {}, "priors": {}}
    new_data["quality"]["sample-rate"] = int(data["srate"])
    new_data["quality"]["lower-frequency"] = {}
    # Factor 0.875 to account for PSD roll off
    new_data["likelihood"]["upper-frequency"] = {ifo: int(0.875 * data["srate"]/2) for ifo in event.meta['interferometers']}
    new_data["quality"]["start-frequency"] = data['f_start']
    new_data["quality"]["segment-length"] = int(data['seglen'])
    new_data["quality"]["window-length"] = int(data['seglen'])
    new_data["quality"]["psd-length"] = int(data['seglen'])

    def decide_fref(freq):
        if (freq >= 5) and (freq < 10):
            return 5
        else:
            return floor(freq/10)*10

    new_data["quality"]["reference-frequency"] = decide_fref(data['f_ref'])

    new_data["priors"]["amp order"] = data['amp_order']
    new_data["priors"]["chirp-mass"] = [data["chirpmass_min"], data["chirpmass_max"]]


    update(event.meta, new_data)
    ledger.update_event(event)


#@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
#@olivaw.command()
@click.argument("event")
@event.command()
def checkifo(event):
    event = ledger.get_event(event)
    if "event time" not in event.event_object.meta:
        print(f"Time not found {event.event_object.name}")
    time = event.event_object.meta['event time']
    gpsstart=time-600
    gpsend=time+600
    bits = ['Bit 0', 'Bit 1', 'Bit 2']

    active_ifo = []
    for ifo in ["L1", "H1", "V1"]:
        frametypes = event.event_object.meta['data']['frame-types']
        urls = find_urls(site=f"{ifo[0]}", frametype=frametypes[ifo], gpsstart=gpsstart, gpsend=gpsend)
        datacache = Cache.from_urls(urls)
        if len(datacache) == 0:
            print(f"No {ifo} data found.")
            continue

        if "state vector" in event.meta:
            state_vector_channel = event.meta['state vector']
        else:                
            state_vector_channel = ast.literal_eval(config.get("data", "state-vector"))

        state = gwpy.timeseries.StateVector.read(
            datacache, state_vector_channel[ifo], start=gpsstart, end=gpsend,
            pad=0  # padding data so that errors are not raised even if found data are not continuous.
        )
        if not np.issubdtype(state.dtype, np.unsignedinteger):
            # if data are not unsigned integers, cast to them now so that
            # we can determine the bit content for the flags
            state = state.astype(
                "uint32",
                casting="unsafe",
                subok=True,
                copy=False,
            )
        flags = state.to_dqflags()

        segments = flags[bits[0]].active
        for bit in bits:
            segments -= ~flags[bit].active

        if len(segments)>0: active_ifo += [ifo]
    click.echo(event.name)
    if event.meta['interferometers'] != active_ifo:
        print(f"Gitlab data\t{event.meta['interferometers']}")
        print(f"Recommended IFOS\t{active_ifo}")

    event.meta['interferometers'] = active_ifo
    ledger.update_event(event)



@click.option("--calibration", "calibration", multiple=True, default=[None], 
              help="The location of the calibration files.")
@click.argument("event")
@event.command()
def calibration(event, calibration):
    event = ledger.get_event(event)
    try:
        event._check_calibration()
    except DescriptionException:
        print(event.title)
        time = event.meta['event time'] 
        if not calibration[0]:
            calibrations = find_calibrations(time)
        else:
            calibrations = {}
            for cal in calibration:
                calibrations[cal.split(":")[0]] = cal.split(":")[1]
        print(calibrations)
        for ifo, envelope in calibrations.items():
            description = f"Added calibration {envelope} for {ifo}."
            try:
                event.repository.add_file(envelope, f"C01_offline/calibration/{ifo}.dat", 
                                                       commit_message=description)
            except GitCommandError as e:
                if "nothing to commit," in e.stderr:
                    pass
            calibrations[ifo] = f"C01_offline/calibration/{ifo}.dat"
        event.meta['calibration'] = calibrations
        ledger.update_event(event)


# @click.argument("data")
# @click.argument("event")
# @event.command()
# def load(event, data):
#     event = ledger.get_event(event)

#     with open(data, "r") as datafile:
#         data = yaml.safe_load(datafile.read())
        
#         event.meta = update(event.meta, data)

#     ledger.update_event(event)



""" 
Olivaw management commands
"""
import os

import pathlib
import click

from asimov import config, logger
from asimov import current_ledger as ledger
from asimov.pipelines import known_pipelines

from asimov import gitlab
from asimov.event import Event, DescriptionException, Production
from asimov.pipeline import PipelineException

@click.group(chain=True)
def manage():
    """Perform management tasks such as job building and submission.
    """
    pass


@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@click.option("--dryrun", "-d", "dryrun", is_flag=True, default=False, help="Print all commands which will be executed without running them")
@manage.command()
def build(event, dryrun):
    """
    Create the run configuration files for a given event for jobs which are ready to run.
    If no event is specified then all of the events will be processed.
    """
    for event in ledger.get_event(event):
        click.echo(f"● Working on {event.name}")
        ready_productions = event.get_all_latest()
        for production in ready_productions:
            click.echo(f"\tWorking on production {production.name}")
            if production.status in {"running", "stuck", "wait", "finished", "uploaded", "cancelled", "stopped"}: continue
            try:
                configuration = production.get_configuration()
            except ValueError:
                try:

                    # if production.rundir:
                    #     path = pathlib.Path(production.rundir)
                    # else:
                    #     path = pathlib.Path(config.get("general", "rundir_default"))
                    
                    if dryrun:
                        print(f"Will create {production.name}.ini")
                    else:
                        #path.mkdir(parents=True, exist_ok=True)
                        config_loc = os.path.join(f"{production.name}.ini")
                        production.make_config(config_loc, dryrun=dryrun)
                        click.echo(f"Production config {production.name} created.")
                        logger.info("Run configuration created.", production=production)
                        try:
                            event.repository.add_file(config_loc,
                                                      os.path.join(f"{production.category}",
                                                                   f"{production.name}.ini"))
                            logger.info("Configuration committed to event repository.",
                                        production=production)
                            ledger.update_event(event)
                            
                        except Exception as e:
                            logger.error(f"Configuration could not be committed to repository.\n{e}",
                                         production=production)
                        os.remove(config_loc)

                except DescriptionException as e:
                    logger.error("Run configuration failed", production=production, channels=["file", "mattermost"])


@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@click.option("--update", "update", default=False, help="Force the git repos to be pulled before submission occurs.")
@click.option("--dryrun", "-d", "dryrun", is_flag=True, default=False, help="Print all commands which will be executed without running them")
@manage.command()
def submit(event, update, dryrun):
    """
    Submit the run configuration files for a given event for jobs which are ready to run.
    If no event is specified then all of the events will be processed.
    """
    for event in ledger.get_event(event):
        ready_productions = event.get_all_latest()
        for production in ready_productions:
            if production.status.lower() in {"running", "stuck", "wait", "processing", "uploaded", "finished", "manual", "cancelled", "stopped"}: continue
            if production.status.lower() == "restart":
                if production.pipeline.lower() in known_pipelines:
                    pipe = known_pipelines[production.pipeline.lower()](production, "C01_offline")
                    pipe.clean(dryrun=dryrun)
                    pipe.submit_dag(dryrun=dryrun)
            else:
                #if production.pipeline.lower() in known_pipelines:
                pipe = production.pipeline#known_pipelines[production.pipeline.lower()](production, "C01_offline")
                try:
                    pipe.build_dag(dryrun=dryrun)
                except PipelineException:
                    logger.error("The pipeline failed to build a DAG file.",
                                 production=production)
                    click.echo(click.style("●", fg="red") + f" Unable to submit {production.name}")
                try:
                    pipe.submit_dag(dryrun=dryrun)
                    click.echo(click.style("●", fg="green") + f" Submitted {production.event.name}/{production.name}")
                    production.status = "running"

                except PipelineException as e:
                    production.status = "stuck"
                    click.echo(click.style("●", fg="red") + f" Unable to submit {production.name}")
                    ledger.update_event(event)
                    logger.error(f"The pipeline failed to submit the DAG file to the cluster. {e}",
                                 production=production)
                # Update the ledger
                ledger.update_event(event)


@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@click.option("--update", "update", default=False, help="Force the git repos to be pulled before submission occurs.")
@manage.command()
def results(event, update):
    """
    Find all available results for a given event.
    """
    for event in ledger.get_event(event):
        click.secho(f"{event.name}")
        for production in event.productions:
            click.echo(f"\t- {production.name}")
            try:
                for result, meta in production.results().items():
                    click.echo(f"- {production.event.name}/{production.name}/{result}, {production.results(result)}")
            except:
                click.echo("\t  (No results available)")
            # print(production.results())

@click.option("--event", "event", default=None, help="The event which the ledger should be returned for, optional.")
@click.option("--update", "update", default=False, help="Force the git repos to be pulled before submission occurs.")
@click.option("--root", "root")
@manage.command()
def resultslinks(event, update, root):
    """
    Find all available results for a given event.
    """
    for event in current_ledger.get_event(event):
        click.secho(f"{event.name}")
        for production in event.productions:
            try:
                for result, meta in production.results().items():
                    print(f"{production.event.name}/{production.name}/{result}, {production.results(result)}")
                    pathlib.Path(os.path.join(root, production.event.name, production.name)).mkdir(parents=True, exist_ok=True)
                    os.symlink(f"{production.results(result)}", f"{root}/{production.event.name}/{production.name}/{result.split('/')[-1]}")
            except AttributeError:
                pass
            # print(production.results())



"""BayesWave Pipeline specification."""


import os
import re
import glob
import subprocess
from ..pipeline import Pipeline, PipelineException, PipelineLogger
from ..ini import RunConfiguration
from ..git import AsimovFileNotFound
from ..storage import Store, AlreadyPresentException

from asimov.utils import set_directory
from asimov import logger
from asimov import config
from shutil import copyfile, copytree
import numpy as np

class BayesWave(Pipeline):
    """
    The BayesWave Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """
    name = "BayesWave"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(BayesWave, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == "bayeswave":
            raise PipelineException

        try:
            self.category = config.get("general", "category")
        except:
            self.category = "C01_offline"
            self.logger.info("Assuming C01_offline calibration.")

    def build_dag(self, user=None, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using bayeswave_pipe

        Parameters
        ----------
        production : str
           The production name.
        user : str
           The user accounting tag which should be used to run the job.

        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """
        if self.production.event.repository:    
            try:
                gps_file = self.production.get_timefile()
            except AsimovFileNotFound:
                if "event time" in self.production.meta:
                    gps_time = self.production.get_meta("event time")
                    with set_directory(os.path.join(self.production.event.repository.directory, self.category)):
                        with open("gpstime.txt", "w") as f:
                            f.write(str(gps_time))
                            gps_file = os.path.join(f"{self.production.category}", f"gpstime.txt")
                            self.production.event.repository.add_file(f"gpstime.txt", gps_file)
                else:
                    raise PipelineException("Cannot find the event time.")
        else:
            gps_time = self.production.get_meta("event time")
            with open("gpstime.txt", "w") as f:
                f.write(str(gps_time))
                gps_file = os.path.join("gpstime.txt")

        

        if self.production.event.repository:
            ini = self.production.get_configuration()
            if not user:
                if self.production.get_meta("user"):
                    user = self.production.get_meta("user")
                else:
                    user = ini._get_user()

            ini.update_accounting(user)

            if 'queue' in self.production.meta:
                queue = self.production.meta['queue']
            else:
                queue = 'Priority_PE'

            ini.set_queue(queue)

            ini.save()

            ini = ini.ini_loc

        else:
            ini = f"{self.production.name}.ini"



        if self.production.rundir:
            rundir = self.production.rundir
        else:
            rundir = os.path.join(config.get("general", "rundir_default"),
                                  self.production.event.name,
                                  self.production.name)
            self.production.rundir = rundir

        try:
            pathlib.Path(directory).mkdir(parents=True, exist_ok=False)
        except:
            pass

        gps_time = self.production.get_meta("event time")

        
        pipe_cmd = os.path.join(config.get("pipelines", "environment"), "bin", "bayeswave_pipe")

        command = [pipe_cmd,
                   #"-l", f"{gps_file}",
                   f"--trigger-time={gps_time}",
                   "-r", self.production.rundir,
                   ini
        ]
            
        self.logger.info(" ".join(command))

        if dryrun:
            print(" ".join(command))
        else:
        
            pipe = subprocess.Popen(command, 
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            out, err = pipe.communicate()
            if "To submit:" not in str(out):
                self.production.status = "stuck"

                raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                            production=self.production.name)
            else:
                if hasattr(self.production.event, "issue_object"):
                    return PipelineLogger(message=out,
                                          issue=self.production.event.issue_object,
                                          production=self.production.name)
                else:
                    return PipelineLogger(message=out,
                                          production=self.production.name)

    def detect_completion(self):
        psds = self.collect_assets()['psds']
        if len(list(psds.values())) > 0:
            return True
        else:
            self.logger.info("Bayeswave job completion was not detected.")
            return False
            
    def after_completion(self):

        try:
            self.collect_pages()
        except FileNotFoundError:
            PipelineLogger(message=b"Failed to copy megaplot pages.",
                           production=self.production.name)
        
        try:
            self.collect_assets()
            self.store_assets()
        except Exception as e:
            PipelineLogger(message=b"Failed to store PSDs.",
                           issue=self.production.event.issue_object,
                           production=self.production.name)
            
        if "supress" in self.production.meta['quality']:
            for ifo in self.production.meta['quality']['supress']:
                if ifo in self.production.meta['interferometers']:
                    self.supress_psd(ifo, 
                                     self.production.meta['quality']['supress'][ifo]['lower'],
                                     self.production.meta['quality']['supress'][ifo]['upper'])
        self.production.status = "uploaded"

        
        
    def before_submit(self):
        pass


    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        category : str, optional
           The category of the job.
           Defaults to "C01_offline".
        production : str
           The production name.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """
        with set_directory(self.production.rundir):

            self.before_submit()

            command = ["condor_submit_dag",
                       "-batch-name", f"bwave/{self.production.event.name}/{self.production.name}",
                       f"{self.production.name}.dag"]

            if dryrun:
                print(" ".join(command))
            else:
                try:
                    dagman = subprocess.Popen(command,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT)
                except FileNotFoundError as error:
                    raise PipelineException("It looks like condor isn't installed on this system.\n"
                                            f"""I wanted to run {" ".join(command)}.""")

                stdout, stderr = dagman.communicate()

                if "submitted to cluster" in str(stdout):
                    cluster = re.search("submitted to cluster ([\d]+)", str(stdout)).groups()[0]
                    self.production.status = "running"
                    self.production.job_id = int(cluster)
                    return int(cluster), PipelineLogger(stdout)
                else:
                    logger.info(stdout)
                    logger.error(stderr)
                    raise PipelineException(f"The DAG file could not be submitted.\n\n{stdout}\n\n{stderr}",
                                            issue=self.production.event.issue_object,
                                            production=self.production.name)
    
    def upload_assets(self):
        """
        Upload the PSDs from this job.
        """
        psds = {}
        detectors = self.production.meta['interferometers']
        sample = self.production.meta["likelihood"]["sample rate"]
        git_location = os.path.join(self.category, "psds")
        
        for detector, asset in self.collect_assets()['psds']:
            self.production.event.repository.add_file(
                asset,
                os.path.join(git_location, str(sample), f"psd_{det}.dat"),
                commit_message = f"Added the PSD for {det}.")

    def store_assets(self):
        """
        Add the assets to the store.
        """
        
        sample_rate = self.production.meta['quality']['sample-rate']
        for detector, asset in self.collect_assets()['psds']:
            store = Store(root=config.get("storage", "directory"))
            try:
                store.add_file(self.production.event.name, self.production.name,
                               file = f"{det}-{sample_rate}-psd.dat")
            except Exception as e:
                error = PipelineLogger(f"There was a problem committing the PSD for {det} to the store.\n\n{e}",
                                       issue=self.production.event.issue_object,
                                       production=self.production.name)
                error.submit_comment()
                
    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and 
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/logs/*.err") + glob.glob(f"{self.production.rundir}/*.err")
        messages = {}
        for log in logs:
            with open(log, "r") as log_f:
                message = log_f.read()
                messages[log.split("/")[-1]] = message
        return messages
                
        
    def collect_assets(self):
        """
        Collect the assets for this job and commit them to the event repository.
        Since this job also generates the PSDs these should be added to the production ledger.
        """
        results_dir = glob.glob(f"{self.production.rundir}/trigtime_*")[0]
        psds = {}
        for det in self.production.meta['interferometers']:
            asset = os.path.join(results_dir, "post", "clean", f"glitch_median_PSD_forLI_{det}.dat")
            if os.path.exists(asset):
                psds[det] = asset

        outputs = {}
        outputs['psds'] = psds
        return outputs
            
    def supress_psd(self, ifo, fmin, fmax):
        """
        Suppress portions of a PSD.
        Author: Carl-Johan Haster - August 2020
        (Updated for asimov by Daniel Williams - November 2020
        """
        store = Store(root=config.get("storage", "directory"))
        sample_rate = self.production.meta['quality']['sample-rate']
        orig_PSD_file = np.genfromtxt(os.path.join(self.production.event.repository.directory, self.category, "psds", str(sample_rate), f"{ifo}-psd.dat"))

        freq = orig_PSD_file[:,0]
        PSD = orig_PSD_file[:,1]

        suppression_region = np.logical_and(np.greater_equal(freq, fmin), np.less_equal(freq, fmax))

        #Suppress the PSD in this region

        PSD[suppression_region] = 1.

        new_PSD = np.vstack([freq, PSD]).T

        asset = f"{ifo}-psd.dat"
        np.savetxt(asset, new_PSD, fmt='%+.5e')
        
        destination = os.path.join(self.category, "psds", str(sample_rate), f"{ifo}-psd.dat")

        try:
            self.production.event.repository.add_file(asset, destination)#, message=f"Added the supresed {ifo} PSD")
        except Exception as e:
            raise PipelineException(f"There was a problem committing the suppresed PSD for {ifo} to the repository.\n\n{e}",
                                    issue=self.production.event.issue_object,
                                    production=self.production.name)
        copyfile(asset, f"{ifo}-{sample_rate}-psd-suppresed.dat")
        try:
            store.add_file(self.production.event.name, self.production.name,
                       file = f"{ifo}-{sample_rate}-psd-suppresed.dat")
        except AlreadyPresentException:
            pass
        self.logger.info(f"PSD supression applied to {ifo} between {fmin} and {fmax}")

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        count = len(glob.glob(os.path.join(self.production.rundir, "*.dag.rescue*")))
        
        if (count < 5) and (count > 0):
            self.submit_dag()
            self.logger.info(f"Bayeswave job was resurrected for the {count} time.")
        else:
            self.logger.error("Bayeswave resurrection not completed as there have already been 5 attempts")

    def html(self):
        """Return the HTML representation of this pipeline."""
        pages_dir = os.path.join(self.production.event.name, self.production.name)
        out = ""
        if self.production.status in {"finished", "uploaded"}:
            out += """<div class="asimov-pipeline">"""
            out += f"""<p><a href="{pages_dir}/index.html">Full Megaplot output</a></p>"""
            out += f"""<img height=200 src="{pages_dir}/plots/clean_whitened_residual_histograms.png"</src>"""
            
            out += """</div>"""
            
        return out

    def collect_pages(self):
        """Collect the HTML output of the pipeline."""
        results_dir = glob.glob(f"{self.production.rundir}/trigtime_*")[0]
        pages_dir = os.path.join(config.get("general", "webroot"), self.production.event.name, self.production.name)
        os.makedirs(pages_dir, exist_ok=True)
        copyfile(os.path.join(results_dir, "index.html"), os.path.join(pages_dir, "index.html"))
        copytree(os.path.join(results_dir, "html"), os.path.join(pages_dir, "html"), dirs_exist_ok=True)
        copytree(os.path.join(results_dir, "plots"), os.path.join(pages_dir, "plots"), dirs_exist_ok=True)



import os
import collections
import yaml
import json
from math import floor
import ligo.gracedb
import gwpy
import pprint
from copy import copy

import click

from git import GitCommandError

from asimov import config
from asimov import current_ledger as ledger
from asimov.storage import Store
from asimov.event import Production, Event,  DescriptionException
from asimov import gitlab

@click.group()
def production():
    """
    Commands to handle productions.
    """
    pass


@click.argument("pipeline")
@click.argument("event")
@click.option("--family", "family", default="Prod", help="The family name of the production, e.g. `prod`.")
@click.option("--comment", "comment", default=None, help="A comment to attach to the production")
@click.option("--needs", "needs", multiple=True, default=[], help="A list of productions which are requirements")
@click.option("--template", "template", default=None, help="The configuration template for the production.")
@click.option("--status", "status", default="ready", help="The initial status of the production.")
@click.option("--approximant", "approximant", default=None, help="The waveform approximant to be used in the production.")
@production.command(help="Add a new production to an event")
def create(event, pipeline, family, comment, needs, template, status, approximant):
    """
    Add a new production to an event.

    """

    event = ledger.get_event(event)[0]
    #
    event_prods = event.productions
    names = [production.name for production in event_prods]
    family_entries = [int(name.split(family)[1]) for name in names if family in name]
    #
    if "bayeswave" in needs:
        bw_entries = [production.name for production in event_prods 
                      if ("bayeswave" in production.pipeline.lower())
                      and (production.review.status not in {"REJECTED", "DEPRECATED"})]
        needs = bw_entries
    #
    production = {"comment": comment, "pipeline": pipeline, "status": status}
    if needs:
        production['needs'] = needs
    if template:
        production['template'] = template
    if approximant:
        production['approximant'] = approximant
    if len(family_entries)>0:
        number = max(family_entries)+1
    else:
        number = 0
    production_dict = {f"{family}{number}": production}
    production = Production.from_dict(production_dict, event=event)
    #
    event.add_production(production)
    click.secho(f"Production added to {event.name}")
    click.echo(f"\t{production}")

    if config.get("ledger", "engine") == "gitlab":
        gitlab_event[0].update_data()
    elif config.get("ledger", "engine") == "yamlfile":
        ledger.events[event.name] = event.to_dict()
        ledger.save()


@click.option("--status", "-s", default=None, help="Change the run status. Options: ready | wait | stuck | stop | cancelled" )
@click.argument("production")
@click.argument("event")
@production.command()
def set(event, production, status):
    """
    Set or update properties such as the status of a production.
    Note that some properties cannot be updated once the production is being run, 
    and you should create a new production instead.
    """
    event = ledger.get_event(event=event)[0]
    production = [production_o for production_o in event.productions if production_o.name == production][0]
    
    accepted_states = {"ready", "wait", "stuck", "stop", "cancelled"}
    if status:
        if status in accepted_states:
            production.status = status
            #ledger.save()
            click.echo(click.style("●", fg="green") + f" {production.name} status updated to {status}")
        else:
            click.echo(click.style("●", fg="red") + f" Unable to change the state of {production.name} to {status}.")


@click.argument("production")
@click.argument("event")
@production.command()
def show(event, production):
    """
    Show the entire metadata for a given production.
    """
    event = ledger.get_event(event=event)[0]
    production = [production_o for production_o in event.productions if production_o.name == production][0]

    meta = copy(production.meta)
    meta.pop("ledger")
    output = yaml.dump(meta, default_flow_style=False)
    click.echo_via_pager(output)
    
@click.option("--file", "file", default=None)
@click.argument("production")
@click.argument("event")
@production.command()
def results(event, production, file, hash=None):
    """
    Fetch or list the results of a production.
    """
    event = ledger.get_event(event)[0]
    production = [production_o for production_o in event.productions if production_o.name == production][0]
    store = Store(root=config.get("storage", "directory"))

    if not file:
        try:
            items = store.manifest.list_resources(event.name, production.name).items()
            click.secho(f"{'Resource':30} {'Hash':32} {'UUID':32}")
            click.secho("-"*96)
            for resource, details in items:
                click.secho(f"{resource:30} {details['hash']:32} {details['uuid']:32}")
        except KeyError:
            click.secho("There are no results for this production.")
    else:
        try:
            click.echo(store.fetch_file(event, production, file, hash))
        except FileNotFoundError:
            click.secho(f"{file} could not be found for this production.")




"""
Logging endpoints
=================


Blueprints for the logging interface.
"""
import datetime

from flask import Blueprint, jsonify, make_response


from flask import Flask
from flask import request

# from asimov.server import database
from asimov import logger

bp = Blueprint("logging", __name__, url_prefix="/log")


@bp.route("/", methods=["POST"])
def add_logging_event():
    """Add a new logging event to the database

    Examples
    --------

    .. http:example:: curl wget python-requests

       POST /log HTTP/1.1
       Content-Type: application/json

       {
         "event":"GW150914",
         "level":"update",
         "message":"This is a test message.",
         "module":"sampling",
         "pipeline":"pycbc",
         "production":"ProdX10"
       }


       HTTP/1.1 201 OK
       Content-Type: application/json
    
       {}
    """
    if request.is_json:
        req = request.get_json()

    entry = logger.log(pipeline=req['pipeline'],
                       module=req['module'],
                       level=req['level'],
                       time=datetime.datetime.now(),
                       event=req['event'],
                       production=req['production'],
                       message=req['message']
                       )
    return make_response({}, 201)


@bp.route("/", methods=["GET"])
def list_logging_events():
    """List all of the log events which match filters given as GET arguments

    Examples
    --------
    .. http:example:: curl wget python-requests

       GET /log HTTP/1.1
       Content-Type: application/json
       
       :query pipeline: RIFT

       
       HTTP/1.1 200 OK
       Content-Type: application/json

       [
         {
           "level": "debug", 
           "message": "Looking-up logs with filterings", 
           "module": "asimov"
         }, 
         {
           "level": "debug", 
           "message": "Looking-up logs with filterings", 
           "module": "asimov"
         }, 
         {
           "level": "debug", 
           "message": "Looking-up logs with filterings", 
           "module": "asimov"
         }
       ]
    """
    filters = request.args
    logger.log(level="debug", message=f"Looking-up logs with filterings: {filters}")
    events = logger.list(**filters)
    return make_response(jsonify([event.__dictrepr__() for event in events]), 200)



"""
Basic endpoints for asimov
==========================

This module implements endpoints which don't handle data, but instead provide things like returning the current version data for things like asimov and the various pipelines.

"""

from flask import Blueprint, request, jsonify, make_response

import asimov

bp = Blueprint("asimov", __name__, url_prefix="")

@bp.route("/", methods=['GET'])
def splash():
    """
    The root endpoint for asimov.

    Example
    -------

    .. http:example:: curl wget python-requests

       GET / HTTP/1.1
       Content-Type: application/json


       HTTP/1.1 200 OK
       Content-Type: application/json

       "asimov"
    """
    return make_response(jsonify(asimov.__name__), 200)

@bp.route("/version", methods=['GET'])
def version():
    """
    Show the version information for asimov and the pipelines it is designed to run.

    Example
    -------

    .. http:example:: curl wget python-requests

       GET /version HTTP/1.1
       Content-Type: application/json


       HTTP/1.1 200 OK
       Content-Type: application/json

       "v0.3.4"
    """
    return make_response(jsonify(asimov.__version__), 200)





"""
Provide the command line interface components for asimov.
"""

import glob

import numpy as np

from asimov import config
from asimov import gitlab

from asimov.utils import find_calibrations
from asimov.pipelines import known_pipelines


frametypes= {"L1": "L1_HOFT_CLEAN_SUB60HZ_C01",
             "H1": "H1_HOFT_CLEAN_SUB60HZ_C01",
             "V1": "V1Online"}

CALIBRATION_NOTE = """
## Calibration envelopes
The following calibration envelopes have been found.
```yaml
---
{}
---
```
"""


ACTIVE_STATES = {"ready", "running", "stuck", "finished", "processing", "stop"}


def add_data(event, yaml_data, json_data=None):
    server, repository = connect_gitlab()
    gitlab_event = gitlab.find_events(repository, subset=event)
    def update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

   
    if yaml_data:
        with open(yaml_data, "r") as datafile:
            data = yaml.safe_load(datafile.read())

        gitlab_event[0].event_object.meta = update(gitlab_event[0].event_object.meta, data)
        gitlab_event[0].update_data()
        print(gitlab_event[0].event_object.meta)

        



"""LALInference Pipeline specification."""


import os
import glob
import re
import subprocess
from ..pipeline import Pipeline, PipelineException, PipelineLogger
from ..ini import RunConfiguration
from asimov import config, logger
from asimov import logging
from asimov.utils import set_directory

import htcondor

class LALInference(Pipeline):
    """
    The LALInference Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """
    name = "lalinference"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(LALInference, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == "lalinference":
            raise PipelineException

    def detect_completion(self):
        """
        Check for the production of the posterior file to signal that the job has completed.
        """
        results_dir = glob.glob(f"{self.production.rundir}/posterior_samples")
        if len(results_dir)>0:
            if len(glob.glob(os.path.join(results_dir[0], f"posterior_*.hdf5"))) > 0:
                return True
            else:
                return False
        else:
            return False
                
    def build_dag(self, psds=None, user=None, clobber_psd=False, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using LALInferencePipe.

        Parameters
        ----------
        production : str
           The production name.
        psds : dict, optional
           The PSDs which should be used for this DAG. If no PSDs are
           provided the PSD files specified in the ini file will be used
           instead.
        user : str
           The user accounting tag which should be used to run the job.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.

        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """

        # Change to the location of the ini file.
        with set_directory(os.path.join(self.production.event.repository.directory,
                              self.category)):
            gps_file = self.production.get_timefile()

            if self.production.rundir:
                rundir = self.production.rundir
            else:
                rundir = os.path.join(os.path.expanduser("~"),
                                      self.production.event.name,
                                      self.production.name)
                self.production.rundir = rundir

            #os.mkdir(self.production.rundir, exist_ok=True)
            ini = f"{self.production.name}.ini"
            command = [
                os.path.join(config.get("pipelines", "environment"),
                "bin",
                "lalinference_pipe"),
                       "-g", f"{gps_file}",
                       "-r", self.production.rundir,
                       ini
            ]

            if dryrun:
                print(" ".join(command))
            else:
                self.logger.info(" ".join(command))
                pipe = subprocess.Popen(command, 
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT)
                out, err = pipe.communicate()
                if err or "Successfully created DAG file." not in str(out):
                    self.production.status = "stuck"
                    if hasattr(self.production.event, "issue_object"):
                        self.logger.error(f"DAG file could not be created.\n{command}\n{out}\n\n{err}")
                        raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                                    issue=self.production.event.issue_object,
                                                    production=self.production.name)
                    else:
                        self.logger.error(f"DAG file could not be created.\n{command}\n{out}\n\n{err}")
                        raise PipelineException(f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                                                production=self.production.name)
                else:
                    self.logger.info("DAG created")
                    if hasattr(self.production.event, "issue_object"):
                        return PipelineLogger(message=out,
                                              issue=self.production.event.issue_object,
                                              production=self.production.name)
                    else:
                        return PipelineLogger(message=out,
                                              production=self.production.name)

    def samples(self):
        """
        Collect the combined samples file for PESummary.
        """
        return glob.glob(os.path.join(self.production.rundir, "posterior_samples", "posterior*.hdf5"))
    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and 
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/log/*.err") + glob.glob(f"{self.production.rundir}/*.err")
        messages = {}
        for log in logs:
            with open(log, "r") as log_f:
                message = log_f.read()
                messages[log.split("/")[-1]] = message
        return messages

    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        category : str, optional
           The category of the job.
           Defaults to "C01_offline".
        production : str
           The production name.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.


        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """

        with set_directory(self.production.rundir):

            self.before_submit(dryrun=dryrun)

            try:
                command = ["condor_submit_dag",
                           "-batch-name", f"lalinf/{self.production.event.name}/{self.production.name}",
                                       os.path.join(self.production.rundir, f"multidag.dag")]

                if dryrun:
                    print(" ".join(command))
                else:
                    dagman = subprocess.Popen(command,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT)

                    stdout, stderr = dagman.communicate()

                    if "submitted to cluster" in str(stdout):
                        cluster = re.search("submitted to cluster ([\d]+)", str(stdout)).groups()[0]
                        self.production.status = "running"
                        self.production.job_id = cluster
                        return cluster, PipelineLogger(stdout)
                    else:
                        raise PipelineException(f"The DAG file could not be submitted.\n\n{stdout}\n\n{stderr}",
                                                issue=self.production.event.issue_object,
                                                production=self.production.name)

            except FileNotFoundError as error:
                    raise PipelineException("It looks like condor isn't installed on this system.\n"
                                            f"""I wanted to run {" ".join(command)}.""")

            
    def after_completion(self):
        cluster = self.run_pesummary()
        self.production.meta['job id'] = int(cluster)
        self.production.status = "processing"

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        try:
            count = self.production.meta['resurrections']
        except:
            count = 0
        if (count < 5) and (len(glob.glob(os.path.join(self.production.rundir, "submit", "*.rescue*")))>0):
            count +=1
            self.submit_dag()

    @classmethod
    def read_ini(cls, filepath):
        """ 
        Read and parse a bilby configuration file.

        Note that bilby configurations are property files and not compliant ini configs.

        Parameters
        ----------
        filepath: str
           The path to the ini file.
        """

        with open(filepath, "r") as f:
            file_content = f.read()

        config_parser = ConfigParser.RawConfigParser()
        config_parser.read_string(file_content)

        return config_parser




@click.option("--event", "event", default=None, help="The event which will be updated")
@click.option("--json", "json_data", default=None)
@olivaw.command(help="Add data from the configurator.")
def configurator(event, json_data=None):
    """
    Add data from the PEConfigurator script to an event.
    """
    server, repository = connect_gitlab()
    gitlab_event = gitlab.find_events(repository, subset=event)
    def update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    if json_data:
        with open(json_data, "r") as datafile:
            data = json.load(datafile)

    new_data = {"quality": {}, "priors": {}}
    new_data["quality"]["sample-rate"] = data["srate"]
    new_data["quality"]["lower-frequency"] = {}
    new_data["quality"]["upper-frequency"] = int(0.875 * data["srate"]/2)
    new_data["quality"]["start-frequency"] = data['f_start']
    new_data["quality"]["segment-length"] = data['seglen']
    new_data["quality"]["window-length"] = data['seglen']
    new_data["quality"]["psd-length"] = data['seglen']

    def decide_fref(freq):
        if (freq >= 5) and (freq < 10):
            return 5
        else:
            return floor(freq/10)*10

    new_data["quality"]["reference-frequency"] = decide_fref(data['f_ref'])

    new_data["priors"]["amp order"] = data['amp_order']
    new_data["priors"]["chirp-mass"] = [data["chirpmass_min"], data["chirpmass_max"]]

    update(gitlab_event[0].event_object.meta, new_data)
    print(gitlab_event[0].event_object.meta)
    gitlab_event[0].update_data()
        



import click
from asimov import config

@click.group()
def configuration():
    """Group for all of the configuration-related command line stuff."""
    pass

@configuration.command()
def show():
    """Show all configuration values"""
    for section, values in config._sections.items():
        click.echo(click.style(section, fg='black', bg='white'))
        for key, value in values.items():
            click.secho(f"\t{key:30}\t{value}", bold=False)



"""
This module contains tools to log data from runs.
Unlike a conventional python process, processes run by asimov may wish to send information to a number of different locations, and store
information for review purposes.
"""

import logging
import os

from asimov import mattermost

from asimov import config
#from .gitlab import EventIssue

import asimov.database as database

logging.getLogger("werkzeug").setLevel(logging.WARNING)
logging.getLogger("MARKDOWN").setLevel(logging.WARNING)

class AsimovLogger():

    def __init__(self, logfile):
        """
        An object for logging asimov productions.

        Parameters
        ----------
        logfile : str
           The location of the log file to be written.
        """
        self.logfile = logfile
        self.mattermost = mattermost.Mattermost()
        self.file_logger = logging.basicConfig(filename=logfile, level=logging.DEBUG)
        
    def log(self, level, message, event=None, production=None, pipeline=None, module=None, time=None):
        """
        Record a log message.

        Parameters
        ----------
        level : str, {DEBUG, INFO, WARNING, ERROR}
           The level of the logging message.
        event : ``asimov.event.Event``
           The event which this message is for.
        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        logger_levels = {"debug": logging.DEBUG,
                         "info": logging.INFO,
                         "warning": logging.WARNING,
                         "error": logging.ERROR,
                         "update": 9}

        if not production:
            production = ""

        logging.log(logger_levels[level.lower()],
                    message)
            
        
    def info(self, message, production=None, channels="file"):
        """
        Record an information message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("info", message, production, channels)

    def error(self, message, production=None, channels="file"):
        """
        Record an error message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("error", message, production, channels)

    def warning(self, message, production=None, channels="file"):
        """
        Record an warning message.

        Parameters
        ----------

        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        self.log("warning", message, production, channels)

    def list(self, offset=0, length=10, **filters):

        class LogEntry:
            def __init__(self, data):
                data = data.strip().split(":")
                self.level = data[0].lower()
                try:
                    self.module = data[1]
                    if self.module == "root":
                        self.module = "asimov"
                except:
                    self.module = None
                try:
                    self.message = data[2]
                except:
                    self.message = None
            def __dictrepr__(self):
                return dict(level=self.level, module=self.module, message=self.message)

        entries = []
        with open(self.logfile, "r") as filehandle:
            for line in filehandle:
                entries.append(LogEntry(line))


        def apply_filter(entries, parameter, value):
            entries = filter(lambda x: getattr(x, parameter) == value,
                             entries)
            return entries

        if filters:
            for parameter, value in filters.items():
                entries = apply_filter(entries, parameter, value)
            entries = list(entries)

                
        if offset:
            return entries[-int(length)+int(offset):-int(offset)]
        else:
            return entries[-int(length):]


class DatabaseLogger(AsimovLogger):
    """
    This class implements a database-backed log collector.
    """

    def __init__(self):
        self.database = database.Logger
    

    def log(cls, level, message, event=None, production=None, pipeline=None, module=None):
        """
        Record a log message.

        Parameters
        ----------
        level : str, {DEBUG, INFO, WARNING, ERROR}
           The level of the logging message.
        production : ``asimov.event.Production``, optional
           The production this message is for.
        message : str
           The free-form message for the log message.
        channels : list, or str, {"file", "mattermost", "gitlab"}, optional
           The list of places where the log message should be sent.
           Defaults to file only.
        """

        logger_levels = {"debug": logging.DEBUG,
                         "info": logging.INFO,
                         "warning": logging.WARNING,
                         "error": logging.ERROR}
        entry = self.database(pipeline,
                              module,
                              level,
                              datetime.datetime.now(),
                              event,
                              production,
                              message)
        entry.save()

    def list(self, **filters):
        """
        Return all logging events which satisfy a set of ``filters``.

        Examples
        --------
        
        """

        events = self.database.list(**filters)



"""Defines the interface with generic analysis pipelines."""
import os
import subprocess
import glob
import re
import time


import warnings
warnings.filterwarnings("ignore", module="htcondor")


import htcondor

from asimov import config, logger
from .storage import Store
from asimov import logging


class PipelineException(Exception):
    """Exception for pipeline problems."""
    def __init__(self, message, issue=None, production=None):
        super(PipelineException, self).__init__(message)
        self.message = message
        self.issue = issue
        self.production = production

    def __repr__(self):
        text = f"""
An error was detected when assembling the pipeline for a production on this event.
Please fix the error and then remove the `pipeline-error` label from this issue.
<p>
  <details>
     <summary>Click for details of the error</summary>
     <p><b>Production</b>: {self.production}</p>
     <p>{self.message}</p>
  </details>
</p>

- [ ] Resolved
"""
        return text

    def submit_comment(self):
        """
        Submit this exception as a comment on the gitlab
        issue for the event.
        """
        if self.issue:
            self.issue.add_label("pipeline-error", state=False)
            self.issue.add_note(self.__repr__())


class PipelineLogger():
    """Log things for pipelines."""
    def __init__(self, message, issue=None, production=None):
        self.message = message#.decode()
        self.issue = issue
        self.production = production

    def __repr__(self):
        text = f"""
One of the productions ({self.production}) produced a log message.
It is copied below.
<p>
  <details>
     <summary>Click for details of the message</summary>
     <p><b>Production</b>: {self.production}</p>
     <p>{self.message}</p>
  </details>
</p>
"""
        return text

    def submit_comment(self):
        """
        Submit this exception as a comment on the gitlab
        issue for the event.
        """
        if self.issue:
            self.issue.add_note(self.__repr__())


class Pipeline():
    """
    Factory class for pipeline specification.
    """
    name = "Asimov Pipeline"
    
    def __init__(self, production, category=None):
        self.production = production

        self.category = production.category
        self.logger = logger
        
    def __repr__(self):
        return self.name.lower()
        
    def detect_completion(self):
        """
        Check to see if the job has in fact completed.
        """
        pass

    def before_submit(self, dryrun=False):
        """
        Define a hook to run before the DAG file is generated and submitted.

        Note, this method should be over-written in the specific pipeline implementation 
        if required.
        It allows the `dryrun` option to be specified in order to only print the commands
        rather than run them.
        """
        pass

    def after_completion(self):
        """
        Define a hook to run after the DAG has completed execution successfully.

        Note, this method should take no arguments, and should be over-written in the 
        specific pipeline implementation if required.
        """
        self.production.status = "finished"
        self.production.meta.pop('job id')

    def collect_assets(self):
        """
        Add the various analysis assets from the run directory to the git repository.
        """

        assets = self.assets
        repo = self.production.event.repository

        for asset in self.assets:
            repo.add_file(asset[0], asset[1])

    def collect_logs(self):
        return {}

    def run_pesummary(self, dryrun=False):
        """
        Run PESummary on the results of this job.
        """

        psds = self.production.psds
        calibration = [os.path.join(self.production.event.repository.directory, cal) for cal in self.production.meta['data']['calibration'].values()]
        configfile = self.production.event.repository.find_prods(self.production.name, self.category)[0]
        command = [
            "--webdir", os.path.join(config.get('general', 'webroot'), self.production.event.name, self.production.name,  "results"),
            "--labels", self.production.name,
            "--gw",
            "--cosmology", config.get('pesummary', 'cosmology'),
            "--redshift_method", config.get('pesummary', 'redshift'),
            "--nsamples_for_skymap", config.get('pesummary', 'skymap_samples'),
            "--evolve_spins", "True",
            "--multi_process", "4",
            "--approximant", self.production.meta['approximant'],
            "--f_low", str(min(self.production.meta['quality']['minimum frequency'].values())),
            "--f_ref", str(self.production.meta['likelihood']['reference frequency']),
            "--regenerate", "redshift mass_1_source mass_2_source chirp_mass_source total_mass_source final_mass_source final_mass_source_non_evolved radiated_energy",
            "--config", os.path.join(self.production.event.repository.directory, self.category, configfile)]
        # Samples
        command += ["--samples"]
        command += self.samples()
        # Calibration information
        command += ["--calibration"]
        command += calibration
        # PSDs
        command += ["--psd"]
        command += psds.values()
        
        self.logger.info(f"Submitted PE summary run. Command: {config.get('pesummary', 'executable')} {' '.join(command)}", production=self.production, channels=['file'])

        if dryrun:
            print(command)

        submit_description = {
            "executable": config.get("pesummary", "executable"),  
            "arguments": " ".join(command),
            "accounting_group": config.get("pipelines", "accounting"),
            "output": f"{self.production.rundir}/pesummary.out",
            "error": f"{self.production.rundir}/pesummary.err",
            "log": f"{self.production.rundir}/pesummary.log",
            "request_cpus": "4",
            "getenv": "true",
            "batch_name": f"PESummary/{self.production.event.name}/{self.production.name}",
            "request_memory": "8192MB",
            "request_disk": "8192MB",
        }

        if dryrun:
            print("SUBMIT DESCRIPTION")
            print("------------------")
            print(submit_description)

        if not dryrun:
            
            hostname_job = htcondor.Submit(submit_description)

            try:
                # There should really be a specified submit node, and if there is, use it.
                schedulers = htcondor.Collector().locate(htcondor.DaemonTypes.Schedd, config.get("condor", "scheduler"))
                schedd = htcondor.Schedd(schedulers)
            except:
                # If you can't find a specified scheduler, use the first one you find
                schedd = htcondor.Schedd()
            with schedd.transaction() as txn:   
                cluster_id = hostname_job.queue(txn)

        else:
            cluster_id = 0
                
        return cluster_id

    def store_results(self):
        """
        Store the PE Summary results
        """

        files = [f"{self.production.name}_pesummary.dat",
                 "posterior_samples.h5",
                 f"{self.production.name}_skymap.fits"]

        for filename in files:
            results = os.path.join(config.get('general', 'webroot'),
                                   self.production.event.name,
                                   self.production.name,
                                   "results", "samples", filename)
            store = Store(root=config.get("storage", "directory"))
            store.add_file(self.production.event.name, self.production.name,
                           file=results)

    def detect_completion_processing(self):
        files = f"{self.production.name}_pesummary.dat"
        results = os.path.join(config.get('general', 'webroot'),
                                   self.production.event.name,
                                   self.production.name,
                                   "results", "samples", files)
        if os.path.exists(results):
            return True
        else:
            return False
            
    def after_processing(self):
        """
        Run the after processing jobs.
        """
        try:
            self.store_results()
            self.production.status = "uploaded"
        except Exception as e:
            raise ValueError(e)

    def eject_job(self):
        """
        Remove a job from the cluster.
        """
        command = ["condor_rm", f"{self.production.meta['job id']}"]
        try:
            dagman = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        
        except FileNotFoundError as error:
            raise PipelineException("It looks like condor isn't installed on this system.\n"
                                    f"""I wanted to run {" ".join(command)}.""")

        stdout, stderr = dagman.communicate()
        if not stderr:
            time.sleep(20)
            self.production.meta.pop('job id')

    def clean(self):
        """
        Remove all of the artefacts from a job from the working directory.
        """
        pass

    def submit_dag(self):
        """
        Submit the DAG for this pipeline.
        """
        raise NotImplementedError
        
    def resurrect(self):
        pass

    def clean(self):
        pass

    @classmethod
    def read_ini(cls, filepath):
        """ 
        Read and parse a pipeline configuration file.

        Parameters
        ----------
        filepath: str
           The path to the ini file.
        """

        with open(filepath, "r") as f:
            file_content = f.read()

        config_parser = ConfigParser.RawConfigParser()
        config_parser.read_string(file_content)

        return config_parser

    def check_progress(self):
        pass

    def html(self):
        """
        Return an HTML representation of this pipeline object.
        """

        out = ""
        out += """<div class="asimov-pipeline">"""
        out += f"""<p class="asimov-pipeline-name">{self.name}</p>"""
        out += """</div>"""
        
        return out

    def collect_pages(self):
        pass
